package org.javafxdata.datasources.provider;

import java.util.HashMap;
import java.util.Map;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.scene.control.TableColumn;

/**
 *
 * Subclasses of DataSource contain specific logic that parses and returns data
 * coming from a DataSourceReader. Implementations have 2 main responsibilities:
 *
 * <br/> 1. provide the required data to the JavaFX Control (via the getItems()
 * call <br/> 2. provide a getColumns() implementation that return TableColumns
 * with the appropirate CellValueFactories, and a getNamedColumn(String name) in
 * order to return a column with a specific name. 
 * <p> 
 * In case the data that has
 * to be rendered in a row is a javafx bean property, the getColumns() can still
 * be used but the TableColumn can also be directly created using the desired
 * property. 
 * <p> Subclasses of this class will populate the data and create the
 * TableColumn instances. 
 *
 * @author johan
 */
public class DataSource<T> {

    protected ObservableList<T> data = FXCollections.<T>observableArrayList();
    private Map<String, TableColumn<T, ?>> columnsMap = new HashMap<String, TableColumn<T, ?>>();
    private final ObservableList<TableColumn<T, ?>> columns = FXCollections.observableArrayList();
    private boolean columnsCreated = false;

    /**
     * Return the column that matches the provided <code>name</code>
     * Subclasses typically have a constructor that accepts column names as Strings.
     * The subclass specific parsing of the data will also create columns and the
     * name of these columns will match the names provided with the constructor.
     * <p>
     * When this method is called before the columns are build, the
     * <code>buildColumns</code> method is called. Subclasses override this 
     * <code>buildColumns</code> method.
     * 
     * @param name the name of the requested column. It is up to the subclass
     * implementation to determine how a column and a name are matched.
     * @return the TableColumn instance that matches the provided name.
     */
    public TableColumn<T, ?> getNamedColumn(String name) {
        if (!columnsCreated) {
            buildColumns();
            columnsCreated = true;
        }
        return columnsMap.get(name);
    }
    
    /**
     * Return all TableColumn instances that are relevant to this DataSource.
     * When this method is called before the columns are build, the
     * <code>buildColumns</code> method is called. Subclasses override this 
     * <code>buildColumns</code> method.
     * 
     * @return all TableColumn instances relevant to this DataSource.
     */
    public ObservableList<TableColumn<T, ?>> getColumns() {
        if (!columnsCreated) {
            buildColumns();
            columnsCreated = true;
        }
        return columns;
    }

    protected void addColumn(String name, TableColumn<T, ?> col) {
        columnsMap.put(name, col);
        columns.add(col);
    }

    protected void buildColumns() {
    }

    protected void add(T t) {
        data.add(t);
    }

    /**
     * Return all data contained in this DataSource. 
     * 
     * @return all data contained in this DataSource
     */
    public ObservableList<T> getData() {
        return data;
    }
}
