package org.javafxdata.datasources.provider;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.application.Platform;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.concurrent.Service;
import javafx.concurrent.Task;
import javafx.concurrent.Worker;
import javax.xml.bind.JAXB;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.dom.DOMSource;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.JsonParser;
import org.codehaus.jackson.map.MappingJsonFactory;
import org.codehaus.jackson.map.ObjectMapper;
import org.javafxdata.datasources.Format;
import org.javafxdata.datasources.reader.DataSourceReader;
import org.javafxdata.datasources.reader.JavaObjectSource;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author johan
 */
public class ObjectDataSource<T> extends TableProviderBase<T> {

    private final DataSourceReader reader;
    private final String tag;
    private final Class itemClass;
    private Service retriever;
    private final BooleanProperty myRunningProperty = new SimpleBooleanProperty();
    private Map<String, Method> methods = new HashMap<String, Method>();
    private Format format;
    private ObjectProperty<T> singleTarget;
    private boolean noResult;

    public ObjectDataSource(DataSourceReader reader, String tag, Class itemClass) {
        this(reader, tag, itemClass, null);
    }

    public ObjectDataSource(DataSourceReader reader, String tag, Class itemClass, String[] columns) {
        this.data = FXCollections.observableArrayList();
        this.reader = reader;
        this.tag = tag;
        this.itemClass = itemClass;
        this.format = reader.getInputFormat();
        if (columns != null) {
            columnNames.addAll(Arrays.asList(columns));
            buildColumns();
            parseMethods(itemClass);
        }
       

    }

    public void setFormat(Format p) {
        this.format = p;
    }

    public void setResultList(ObservableList<T> myList) {
        this.data = myList;
    }
    
    public void setResultObject (ObjectProperty<T> target) {
        this.singleTarget = target;
    }
    
    public void setNoResult (boolean b) {
        this.noResult = b;
    }

    public ReadOnlyProperty runningProperty() {
        return retriever.runningProperty();
    }

    public ReadOnlyProperty<Worker.State> stateProperty() {
        return retriever.stateProperty();
    }

    public Service retrieve() {
        if (Format.JAVA_OBJECT.equals(format)) {
                try {
                    retriever = new ObjectDataRetriever((JavaObjectSource) reader);
                } catch (IOException ex) {
                    Logger.getLogger(ObjectDataSource.class.getName()).log(Level.SEVERE, null, ex);
                    ex.printStackTrace();
                }
        }
        if (Format.XML.equals(format)) {
                retriever = new XmlDataRetriever(reader);
        }
        if (Format.JSON.equals(format)) { 
                retriever = new JsonDataRetriever(reader);
        }
        retriever.start();
        return retriever;
    }

    private void parseMethods(Class x) {
        Method[] xMethods = x.getMethods();
        for (Method xMethod : xMethods) {
            if (xMethod.getName().startsWith("get")
                    && xMethod.getParameterTypes().length == 0
                    && !void.class.equals(xMethod.getReturnType())) {
                String shortenedMethodName = xMethod.getName().substring(3, 4).toLowerCase() + xMethod.getName().substring(4);
                methods.put(shortenedMethodName, xMethod);
            }
        }
    }

    @Override
    public Object getMyElement(T row, String columnName) {
        if (row != null) {
            Method m = methods.get(columnName);
            if (m != null) {
                try {
                    return m.invoke(row);
                } catch (IllegalAccessException ex) {
                    Logger.getLogger(ObjectDataSource.class.getName()).log(Level.SEVERE, null, ex);
                } catch (IllegalArgumentException ex) {
                    Logger.getLogger(ObjectDataSource.class.getName()).log(Level.SEVERE, null, ex);
                } catch (InvocationTargetException ex) {
                    Logger.getLogger(ObjectDataSource.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
        return null;
    }

    class ObjectDataRetriever extends Service {

        private ObjectInputStream inputStream;

        public ObjectDataRetriever(JavaObjectSource reader) throws IOException {
            this.inputStream = new ObjectInputStream (reader.getInputStream());
        }

        @Override
        protected Task createTask() {
            Task task = new Task() {

                @Override
                protected Object call() throws IOException, ClassNotFoundException {//throws Exception {
                    T entry = (T) inputStream.readObject();
                    while (entry != null) {
                        add(entry);
                        try {
                            entry = (T) inputStream.readObject();
                        } catch (EOFException e) {
                            entry = null;
                        }
                    }
                    return this;
                }
            };
            return task;
        }
    }

    class JsonDataRetriever extends Service {

        final DataSourceReader myReader;

        public JsonDataRetriever(DataSourceReader reader) {
            this.myReader = reader;
            myRunningProperty.bind(this.runningProperty());
        }

        @Override
        protected Task createTask() {
            Task task = new Task() {

                @Override
                protected Object call() throws Exception {
              //      try {
                        InputStream is = myReader.getInputStream();
                        if (singleTarget != null) {
                            try {

                                MappingJsonFactory mappingJsonFactory = new MappingJsonFactory();
                                JsonParser jp = mappingJsonFactory.createJsonParser(is);

                                JsonNode rootNode = jp.readValueAsTree();
                                ObjectMapper om = new ObjectMapper();
                                Object o = om.readValue(rootNode, itemClass);
                                final T entry = (T) o;
                                Platform.runLater(new Runnable() {

                                    @Override
                                    public void run() {
                                        singleTarget.set(entry);

                                    }
                                });
                            } catch (Exception e) {
                                System.out.println("error parsing json");
                                e.printStackTrace();
                            }
                            return this;
                        }
                        JsonParser jp = new MappingJsonFactory().createJsonParser(is);
                        JsonNode rootNode = jp.readValueAsTree();
                        JsonNode cNode = rootNode;
                        if (tag != null) {
                            cNode = rootNode.findValue(tag);
                        }
                        Iterator<JsonNode> iterator = cNode.iterator();

                        while (iterator.hasNext()) {
                            JsonNode candidate = iterator.next();
                            ObjectMapper om = new ObjectMapper();
                            final T entry = (T) om.readValue(candidate, itemClass);

                            Platform.runLater(new Runnable() {

                                @Override
                                public void run() {
                                    add(entry);
                                }
                            });
                        }
//                    } catch (Exception e) {
//                        e.printStackTrace();
//                    }
                    return this;
                }
            };
            return task;
        }
    }

    class XmlDataRetriever extends Service {

        final DataSourceReader myReader;

        public XmlDataRetriever(DataSourceReader reader) {
            this.myReader = reader;
            // TODO: we have a separate myRunningProperty in case the info is requested before this service is created
            myRunningProperty.bind(this.runningProperty());
        }

        @Override
        protected Task createTask() {
            Task task = new Task() {

                @Override
                protected Object call() throws Exception {
                    InputStream is = myReader.getInputStream();
                    if (noResult == true) {
                        is.close();
                        return this;
                    }
                    if (singleTarget != null) {
                        System.out.println("I have to unmarshal to " + itemClass);
                        final T entry = (T) JAXB.unmarshal(is, itemClass);
                        Platform.runLater(new Runnable() {
                            public void run() {
                                singleTarget.set(entry);
                            }
                          }
                        );
                        return this;
                    }
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();

                    DocumentBuilder db = dbf.newDocumentBuilder();
                    Document doc = db.parse(is);
                    NodeList childNodes = doc.getElementsByTagName(tag);
                    int count = childNodes.getLength();
                    for (int i = 0; i < count; i++) {
                        Node node = childNodes.item(i);
                        DOMSource source = new DOMSource(node);
                        final T entry = (T) JAXB.unmarshal(source, itemClass);
                        Platform.runLater(new Runnable() {

                            @Override
                            public void run() {
                                add(entry);
                            }
                        });
                    }
                    return this;
                }
            };
            return task;
        }
    }
}
