package org.javafxdata.datasources.provider;

import java.util.Observable;
import javafx.beans.property.ObjectProperty;
import javafx.beans.value.ObservableValue;
import javafx.collections.ObservableList;
import org.javafxdata.datasources.Format;
import org.javafxdata.datasources.reader.DataSourceReader;

/**
 *
 * @author johan
 */
public class ObjectDataSourceBuilder<T> {

    private DataSourceReader reader;
    private ObservableList<T> existingList;
    private ObjectProperty<T> singleTarget;
    private Class itemClass;
    private String tag;
    private String[] columns;
    private Format format;
    private boolean noResult;
    
    public static <T> ObjectDataSourceBuilder create () {
        return new ObjectDataSourceBuilder<T>();
    }
    
    protected ObjectDataSourceBuilder() {
        
    }
    
    /**
     * Specify the DataSourceReader that contains the inputStream for this
     * ObjectDataSource. Typically, this DataSourceReader is created using
     * a RestRequestBuilder.
     * @param reader the DataSourceReader containing the input for this
     * ObjectDataSource
     * @return the current ObjectDataSourceBuilder
     */
    public ObjectDataSourceBuilder dataSourceReader (DataSourceReader reader) {
        this.reader = reader;
        return this;
    }
    
    
    /**
     * The resulting entities for the ObjectDataSource created by this builder
     * will be added by the supplied ObservableList. If this method is not
     * called, the ObjectDataSource will create an own, empty ObservableList
     * instance.
     * <br/>
     * This method is particularly useful in case the entities obtained via this
     * ObjectDataSource need to be merged with existing entities.
     * @param list the ObservableList that will be populated with the obtained
     * elements. This list is <b>not</b> cleared, hence all existing elements
     * in the list will be kept.
     * @return the current ObjectDataSourceBuilder
     */
    public ObjectDataSourceBuilder resultList (ObservableList<T> list) {
        this.existingList = list;
        return this;
    }
    
    /**
     * In case the result of the data-retrieval is a single Object, an
     * ObjectProperty should be passed here. Once the data is 
     * retrieved, this ObjectProperty will contain the data
     * @param target the ObjectProperty that will contain the retrieved data
     * @return 
     */
    public ObjectDataSourceBuilder resultObject (ObjectProperty<T> target) {
        this.singleTarget = target;
        return this;
    }
    
    public ObjectDataSourceBuilder noResult (boolean b) {
        this.noResult = b;
        return this;
    }
    
    /**
     * Specify the Class defining the entities that should be obtained when
     * parsing the ObjectDataSource.
     * @param clazz a Class containing the structure of the entities. The
     * items in the resulting stream will be parsed into instances of this
     * Class.
     * @return the current ObjectDataSourceBuilder
     */
    public ObjectDataSourceBuilder itemClass (Class clazz) {
        this.itemClass = clazz;
        return this;
    }
    
    /**
     * Specify the tag that separates the entities contained in the 
     * DataSourceReader
     * @param tag
     * @return the current ObjectDataSourceBuilder
     */
    public ObjectDataSourceBuilder itemTag(String tag) {
        this.tag = tag;
        return this;
    }
    
    /**
     * Specify which columns should be taken into account when creating
     * a <code>TableView</code> based on the items in this 
     * ObjectDataSource.
     * <br/>
     * Valid column names are JavaBean properties (e.g. if the class
     * specified via the <code>itemClass<code> method contains a method
     * <code>setFoo</code>, "<code>foo</code>" is a valid column name.
     * @param columns A list of String instances, defining the column names.
     * @return 
     */
    public ObjectDataSourceBuilder columns (String... columns) {
        this.columns = columns;
        return this;
    }
    
    public ObjectDataSourceBuilder format (Format format) {
        this.format = format;
        return this;
    }
    /**
     * Create the ObjectDataSource specified by this builder.
     * @return the resulting ObjectDataSource
     */
    public ObjectDataSource<T> build() {
        if (reader == null) {
            throw new IllegalArgumentException("DataSourceReader not specified");
        }
        ObjectDataSource<T> answer = new ObjectDataSource(reader, tag, itemClass, columns);
        if (existingList != null) {
            answer.setResultList(existingList);
        }
        if (singleTarget != null) {
            answer.setResultObject(singleTarget);
        }
        if (format != null) {
            answer.setFormat(format);
        }
        if (noResult) answer.setNoResult(true);
        return answer;
    }

    
}
