package org.javafxdata.datasources.provider;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.application.Platform;
import javafx.beans.property.ReadOnlyFloatWrapper;
import javafx.beans.property.ReadOnlyLongWrapper;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.beans.property.ReadOnlyStringWrapper;
import javafx.beans.value.ObservableValue;
import javafx.concurrent.Service;
import javafx.concurrent.Task;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableColumn.CellDataFeatures;
import javafx.util.Callback;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.javafxdata.datasources.reader.DataSourceReader;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 *
 * @author johan
 */
public class XMLDataSource extends DataSource<Node> implements ListWrapper, TableWrapper {

    private XPath xpath;
    private List<String> selectedColumnNames = new LinkedList<String>();
    private List<Class> selectedColumnClasses = new LinkedList<Class>();

    public XMLDataSource(DataSourceReader reader, String tagName) throws IOException, ParserConfigurationException, SAXException {
        this(reader, tagName, new String[0]);
    }

    /**
     * Create a new XMLDataSource without specifying the column types. In this case,
     * it is assumed that all columns will contain String properties.
     * @param reader
     * @param tagName
     * @param cols the name of the columns
     * @throws IOException
     * @throws ParserConfigurationException
     * @throws SAXException 
     */
    public XMLDataSource(DataSourceReader reader, String tagName, String... cols) throws IOException, ParserConfigurationException, SAXException {
        this(reader, tagName, null, cols);
    }
    
    /**
     * Create a new XMLDataSource and specify the column types.
     * @param reader
     * @param tagName
     * @param cols the name of the columns
     * @throws IOException
     * @throws ParserConfigurationException
     * @throws SAXException 
     */
    public XMLDataSource(DataSourceReader reader, String tagName, List<Class> clazzes, String... cols) throws IOException, ParserConfigurationException, SAXException {
       // TODO we might use a builder approach here to add (clazz, name) properties
        selectedColumnNames.addAll(Arrays.asList(cols));
        if ((clazzes != null) && (clazzes.size() == cols.length)) {
            selectedColumnClasses.addAll(clazzes);
        }
        else  {
          for (int i = 0; i < cols.length; i++) {
            selectedColumnClasses.add(String.class);
          }
        }
        XPathFactory factory = XPathFactory.newInstance();
        this.xpath = factory.newXPath();
        DataRetriever retriever = new DataRetriever(reader, tagName);
        retriever.start();
    }

    public XMLDataSource(String string, String tagName) throws ParserConfigurationException, SAXException, IOException {
        XPathFactory factory = XPathFactory.newInstance();
        this.xpath = factory.newXPath();

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        initializeDataRows(db.parse(string), tagName);
    }

    public XMLDataSource(InputSource source, String tagName) throws ParserConfigurationException, SAXException, IOException {
        XPathFactory factory = XPathFactory.newInstance();
        this.xpath = factory.newXPath();

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        initializeDataRows(db.parse(source), tagName);
    }

    public XMLDataSource(InputStream stream, String tagName) throws ParserConfigurationException, SAXException, IOException {
        XPathFactory factory = XPathFactory.newInstance();
        this.xpath = factory.newXPath();

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        initializeDataRows(db.parse(stream), tagName);
    }

    public XMLDataSource(File file, String tagName) throws ParserConfigurationException, SAXException, IOException {
        XPathFactory factory = XPathFactory.newInstance();
        this.xpath = factory.newXPath();

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        initializeDataRows(db.parse(file), tagName);
    }

    private void initializeDataRows(Document xml, String tagName) {
        xml.getDocumentElement().normalize();

        NodeList xmlNodes;
        if (tagName != null && !tagName.isEmpty()) {
            xmlNodes = xml.getElementsByTagName(tagName);
        } else {
            xmlNodes = xml.getDocumentElement().getChildNodes();
        }

        for (int i = 0; i < xmlNodes.getLength(); i++) {
            add(xmlNodes.item(i));
        }
    }

    public TableColumn<Node, ?> getNamedColumn(final String name, final Class clazz) {
        TableColumn<Node, ? extends Object> myanswer = super.getNamedColumn(name);
        
        if (myanswer != null) {
            return myanswer;
        } else {
            TableColumn<Node, Object> answer = new TableColumn<Node, Object>(name);

            answer.setCellValueFactory(new Callback<TableColumn.CellDataFeatures<Node, Object>, ObservableValue<Object>>() {

                @Override
                public ObservableValue<Object> call(CellDataFeatures<Node, Object> cdf) {
                    Node myRow = (Node) cdf.getValue();
                    if (clazz.equals(Long.class)) {
                        return (ObservableValue) new ReadOnlyLongWrapper(Long.parseLong(getMyElement(myRow, name)));
                    }
                    if (clazz.equals(Float.class)) {
                        return (ObservableValue) new ReadOnlyFloatWrapper(Float.parseFloat(getMyElement(myRow, name)));
                    } else {
                        return (ObservableValue) new ReadOnlyStringWrapper(getMyElement(myRow, name));
                    }
                }
            });
            return myanswer;
        }
    }
    
    protected Class getColumnClass(String columnName) {
        Class clazz = String.class;
        if (selectedColumnClasses != null) {
            for (int i = 0; i < selectedColumnNames.size(); i++) {
                if (columnName.equals(selectedColumnNames.get(i))) {
                    return selectedColumnClasses.get(i);
                }
            }
        }
        return clazz;
    }
    
    @Override
    protected void buildColumns() {
        for (final String columnName : selectedColumnNames) {
            TableColumn<Node, Object> tc = new TableColumn<Node, Object>(columnName);

            tc.setCellValueFactory(new Callback<TableColumn.CellDataFeatures<Node, Object>, ObservableValue<Object>>() {

               @Override
                public ObservableValue<Object> call(CellDataFeatures<Node, Object> cdf) {
                    Node myRow = cdf.getValue();
                    Class clazz = getColumnClass(columnName);
                    String value = getMyElement(myRow, columnName);
                    if (Integer.class == clazz || Long.class == clazz) {
                        Long l = null;
                        if (value.length() > 0)
                        try {
                            l = Long.parseLong(value);
                        } catch (NumberFormatException e) {
                        }
                        return (ObservableValue) (new ReadOnlyLongWrapper(l));
                    } else if (Double.class == clazz || Float.class == clazz) {
                        Float f = null;
                        if (value.length() > 0)
                        try {
                            f = Float.parseFloat(value);
                        } catch (NumberFormatException e) {
                        }
                        return (ObservableValue) (new ReadOnlyFloatWrapper(f));
                    } else if (Date.class == clazz || Calendar.class == clazz) {
                        SimpleDateFormat df = new SimpleDateFormat();
                        Date date = null;
                        if (value.length() > 0)
                        try {
                            date = df.parse(value);
                        } catch (ParseException ex) {
                            Logger.getLogger(XMLDataSource.class.getName()).log(Level.SEVERE, null, ex);
                        }
                        return (ObservableValue) (new ReadOnlyObjectWrapper<Date>(date));
                    }
                    return (ObservableValue) (new ReadOnlyStringWrapper(value));
                }
              
//                @Override
//                public ObservableValue<Object> call(CellDataFeatures<Node, Object> cdf) {
//                    Node myRow = cdf.getValue();
//                    // TODO: here, we always return a String. Should check on other types.
//                    return (ObservableValue) (new ReadOnlyStringWrapper(getMyElement(myRow, columnName)));
//                }
            });
            super.addColumn(columnName, tc);
        }
    }

    private String getMyElement(Node node, String columnName) {
        if (node != null) {
            try {
                XPathExpression expression = xpath.compile(columnName);
                String e = expression.evaluate(node);
                return e;
            } catch (XPathExpressionException ex) {
                Logger.getLogger(XMLDataSource.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return null;
    }

    class DataRetriever extends Service {

        private final DataSourceReader myReader;
        private final String myTagName;

        public DataRetriever(DataSourceReader reader, String tagName) {
            this.myReader = reader;
            this.myTagName = tagName;
        }

        @Override
        protected Task createTask() {
            Task task = new Task() {

                @Override
                protected Object call() throws Exception {
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    DocumentBuilder db = dbf.newDocumentBuilder();
                    initializeDataRows(db.parse(myReader.getInputStream()), myTagName);
                    Document xml = db.parse(myReader.getInputStream());
                    String tagName = myTagName;
                    xml.getDocumentElement().normalize();

                    NodeList xmlNodes;
                    if (tagName != null && !tagName.isEmpty()) {
                        xmlNodes = xml.getElementsByTagName(tagName);
                    } else {
                        xmlNodes = xml.getDocumentElement().getChildNodes();
                    }

                    for (int i = 0; i < xmlNodes.getLength(); i++) {
                        final Node item = xmlNodes.item(i);
                        Platform.runLater(new Runnable() {

                            @Override
                            public void run() {
                                add(item);
                            }
                        });
                    }
                    return this;
                }
            };

            return task;
        }
    }
}
