package org.javafxdata.datasources.reader;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.security.GeneralSecurityException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.javafxdata.datasources.Format;
import org.javafxdata.datasources.util.OAuth;

/**
 * Implementation of DataSourceReader for obtaining data from a network source
 * (e.g. a static XML file, a dynamically served REST interface,...). Instances
 * of this class can be created and passed in the constructor of a protocol
 * implementations in the org.javafxdata.datasources.io package. <br/> Note that
 * using this class does not create a binding between the physical network
 * source and the view that renders the data.
 *
 * @author johan
 */
public class NetworkSource implements DataSourceReader {
    private InputStream is;
    private String urlBase;
    private final String address;
    private Map<String, String> requestProperties;
    private Map<String, String> queryParams = new HashMap<String, String>();
    private Map<String, String> formParams = new HashMap<String, String>();
    private String dataString;
    private String requestMethod;
    private Format format;
    private String consumerKey;
    private String consumerSecret;
    

    public NetworkSource (String address) {
        this.address = address;
    }
    
    public NetworkSource(String address, Format format) {
        this.address = address;
        this.format = format;
    }

    public void setDataString(String dataString) {
        this.dataString = dataString;
    }

    public void setRequestProperties(Map<String, String> requestProperties) {
        this.requestProperties = requestProperties;
    }

    public void setRequestMethod(String requestMethod) {
        this.requestMethod = requestMethod;
    }

    @Override
    public InputStream getInputStream() throws IOException {
        URL url = new URL(address);
        URLConnection connection = url.openConnection();
        if (consumerKey != null) {
            try {
                Map<String, String> allParams = new HashMap<String, String>();
                allParams.putAll(getQueryParams());
                allParams.putAll(getFormParams());
                String header = OAuth.getHeader(requestMethod, urlBase, allParams, consumerKey, consumerSecret);
                connection.addRequestProperty("Authorization", header);

            } catch (UnsupportedEncodingException ex) {
                Logger.getLogger(NetworkSource.class.getName()).log(Level.SEVERE, null, ex);
            } catch (GeneralSecurityException ex) {
                Logger.getLogger(NetworkSource.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        if (requestMethod != null) {
            ((HttpURLConnection) connection).setRequestMethod(requestMethod);
        }

        if (requestProperties != null) {
            for (Map.Entry<String, String> requestProperty : requestProperties.entrySet()) {
                connection.addRequestProperty(requestProperty.getKey(), requestProperty.getValue());
            }
        }

        if (dataString != null) {
            connection.setDoOutput(true);
            OutputStreamWriter outputStreamWriter = new OutputStreamWriter(connection.getOutputStream());
            outputStreamWriter.write(dataString);
            outputStreamWriter.close();
        }

        is = connection.getInputStream();
        return is;
    }

    @Override
    public Format getInputFormat() {
        return this.format;
    }

    /**
     * @return the consumerKey
     */
    public String getConsumerKey() {
        return consumerKey;
    }

    /**
     * @param consumerKey the consumerKey to set
     */
    public void setConsumerKey(String consumerKey) {
        this.consumerKey = consumerKey;
    }

    /**
     * @return the consumerSecret
     */
    public String getConsumerSecret() {
        return consumerSecret;
    }

    /**
     * @param consumerSecret the consumerSecret to set
     */
    public void setConsumerSecret(String consumerSecret) {
        this.consumerSecret = consumerSecret;
    }

    /**
     * @return the urlBase
     */
    public String getUrl() {
        return urlBase;
    }

    /**
     * @param urlBase the urlBase to set
     */
    public void setUrl(String url) {
        this.urlBase = url;
    }

    /**
     * @return the queryParams
     */
    public Map<String, String> getQueryParams() {
        return queryParams;
    }

    /**
     * @param queryParams the queryParams to set
     */
    public void setQueryParams(Map<String, String> queryParams) {
        this.queryParams = queryParams;
    }

    /**
     * @return the formParams
     */
    public Map<String, String> getFormParams() {
        return formParams;
    }

    /**
     * @param formParams the formParams to set
     */
    public void setFormParams(Map<String, String> formParams) {
        this.formParams = formParams;
    }
}
