package org.javafxdata.datasources.reader;

import java.util.HashMap;
import java.util.Map;
import org.javafxdata.datasources.Format;

/**
 *
 * @author johan
 */
public class RestRequestBuilder {

    private StringBuilder path;
    private final Map<String, String> headers;
    private final Map<String, String> queryParams;
    private final Map<String, String> formParams;
    private StringBuilder queryString;
    private StringBuilder dataString;
    private String method;
    private String consumerKey;
    private String consumerSecret;
    private Format format = Format.XML;

    public RestRequestBuilder(String host) {
        this.path = new StringBuilder(host);
        this.headers = new HashMap<String, String>();
        this.queryParams = new HashMap<String, String>();
        this.formParams = new HashMap<String, String>();
        this.method = "GET";
    }

    public RestRequestBuilder queryParam(String okey, String ovalue) {
        this.queryParams.put(okey, ovalue);
        String key = encodeQueryParam(okey);
        String value = encodeQueryParam(ovalue);
        if (queryString == null) {
            queryString = new StringBuilder(key).append("=").append(value);
        } else {
            queryString.append("&").append(key).append("=").append(value);
        }
        return this;
    }

    public RestRequestBuilder formParam(String key, String value) {
        this.formParams.put(key, value);
        if (dataString == null) {
            dataString = new StringBuilder(key).append("=").append(value);
        } else {
            dataString.append("&").append(key).append("=").append(value);
        }
        return this;
    }

    public RestRequestBuilder path(String path) {
        this.path.append("/").append(path);
        return this;
    }
    
    public RestRequestBuilder consumerKey(String v) {
        this.consumerKey = v;
        return this;
    }
    
    public RestRequestBuilder consumerSecret(String v) {
        this.consumerSecret = v;
        return this;
    }

    /**
     * Set the Accept HTTP header for the request.
     *
     * @param accept
     * @return
     */
    public RestRequestBuilder accept(String accept) {
        this.headers.put("Accept", accept);
        return this;
    }

    /**
     * Set the Content-Type HTTP header for the request.
     *
     * @param accept
     * @return
     */
    public RestRequestBuilder type(String type) {
        this.headers.put("Content-Type", type);
        return this;
    }

    /**
     * Set the HTTP method to use for the request.
     *
     * @param method
     * @return
     */
    public RestRequestBuilder method(String method) {
        this.method = method;
        return this;
    }
    
    /**
     * Set the format of the expected response data
     * @param format
     * @return 
     */
    public RestRequestBuilder format (Format format) {
        this.format = format;
        return this;
    }

    public RestRequest build() {
        String request = path.toString();
        if (queryString != null) {
            request = request + "?" + queryString;
        }
        RestRequest source = new RestRequest (request, format);
        source.setUrl(path.toString());
        source.setQueryParams(queryParams);
        source.setFormParams(formParams);
        if (consumerKey != null) {
            source.setConsumerKey(consumerKey);
        }
        if (consumerSecret != null) {
            source.setConsumerSecret(consumerSecret);
        }
        if (dataString != null) {
            source.setDataString(dataString.toString());
        }
        source.setRequestMethod(method);
        source.setRequestProperties(headers);
        return source;
    }

    private String encodeQueryParam(String src) {
        StringBuilder sb = new StringBuilder();
        int i = 0;
        while (i < src.length()) {
            char c = src.charAt(i);
            if (Character.isAlphabetic(c) || Character.isDigit(c)) {
                sb.append(c);
            } else {
                if (c == ' ') {
                    sb.append('+');
                } else {
                    escapeSpecialChar(sb, c);
                }
            }
            i++;
        }
        return sb.toString();
    }
    private final static char[] HEX = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

    private static void escapeSpecialChar(StringBuilder sb, int b) {
        sb.append('%');
        sb.append(HEX[b >> 4]);
        sb.append(HEX[b & 0x0F]);
    }
}
